0	ICON	termplay.ico
//1	24	termplay.manifest
100	BITMAP	open.bmp
101	BITMAP	rewind.bmp
102	BITMAP	pause.bmp
103	BITMAP	play.bmp
104	BITMAP	selstart.bmp
105	BITMAP	selend.bmp
106	BITMAP	export.bmp
wood1	BITMAP	wood1.bmp
