void get_w_format(const char **format); // get it or die noisily
void get_r_format(const char **format); // uhm... just like above
int open_out(char **file_name, const char *format_ext, int append); // find a name if not provided
